@@ -1,9 +1,5 @@
-import asyncio
-import logging
 import logging.config
 import sys
-from functools import wraps
-from typing import Any, Callable
 
 import click
 import uvicorn
@@ -11,16 +7,6 @@ import uvicorn
 from sagemaker_shim.app import app
 from sagemaker_shim.logging import LOGGING_CONFIG
 
-logger = logging.getLogger(__name__)
-
-
-def cli_coroutine(f: Callable[..., Any]) -> Callable[..., Any]:
-    @wraps(f)
-    def wrapper(*args: Any, **kwargs: Any) -> Any:
-        return asyncio.run(f(*args, **kwargs))
-
-    return wrapper
-
 
 @click.group()
 def cli() -> None:
@@ -30,7 +16,7 @@ def cli() -> None:
 @cli.command(short_help="Start the model server")
 def serve() -> None:
     logging.config.dictConfig(LOGGING_CONFIG)
-    uvicorn.run(app=app, host="0.0.0.0", port=8080, log_config=None)
+    uvicorn.run(app=app, host="0.0.0.0", port=8080, log_config=None, workers=1)
 
 
 if __name__ == "__main__":
