@@ -8,22 +8,26 @@ from base64 import b64decode
 from functools import cached_property
 from importlib.metadata import version
 from pathlib import Path
+from shutil import rmtree
 from tempfile import TemporaryDirectory
 from typing import TYPE_CHECKING, Any
+from zipfile import BadZipFile
 
 import boto3
 from pydantic import BaseModel, validator
 
+from sagemaker_shim.exceptions import ZipExtractionError
 from sagemaker_shim.logging import STDOUT_LEVEL
 from sagemaker_shim.utils import safe_extract
 
 if TYPE_CHECKING:
-    from mypy_boto3_s3 import S3Client
+    from mypy_boto3_s3 import S3Client  # pragma: no cover
 else:
     S3Client = object
 
 
 logger = logging.getLogger(__name__)
+lock = asyncio.Lock()
 
 BUCKET_NAME_REGEX = re.compile(r"^[a-zA-Z0-9.\-_]{1,255}$")
 BUCKET_ARN_REGEX = re.compile(
@@ -31,7 +35,16 @@ BUCKET_ARN_REGEX = re.compile(
 )
 
 
+def validate_bucket_name(v: str) -> str:
+    if BUCKET_NAME_REGEX.match(v) or BUCKET_ARN_REGEX.match(v):
+        return v
+    else:
+        raise ValueError("Invalid bucket name")
+
+
 class InferenceIO(BaseModel):
+    """A single input or output file for an inference job"""
+
     relative_path: Path
     bucket_name: str
     bucket_key: str
@@ -40,13 +53,16 @@ class InferenceIO(BaseModel):
     class Config:
         frozen = True
 
-    # TODO validation of attrs
+    @validator("bucket_name")
+    def validate_bucket_name(cls, v: str) -> str:  # noqa:B902
+        return validate_bucket_name(v)
 
     def local_file(self, path: Path) -> Path:
-        # TODO handle images and non relative_paths
+        """The local location of the file"""
         return path / self.relative_path
 
     def download(self, *, input_path: Path, s3_client: S3Client) -> None:
+        """Download this file from s3"""
         dest_file = self.local_file(path=input_path)
 
         logger.info(
@@ -57,7 +73,6 @@ class InferenceIO(BaseModel):
 
         if self.decompress:
             with TemporaryDirectory() as tmp_dir:
-                # TODO add tests
                 zipfile = Path(tmp_dir) / "src.zip"
                 with zipfile.open("wb") as f:
                     s3_client.download_fileobj(
@@ -65,7 +80,12 @@ class InferenceIO(BaseModel):
                         Key=self.bucket_key,
                         Fileobj=f,
                     )
-                safe_extract(src=zipfile, dest=dest_file.parent)
+                try:
+                    safe_extract(src=zipfile, dest=dest_file.parent)
+                except BadZipFile as error:
+                    raise ZipExtractionError(
+                        "Input zip file could not be extracted"
+                    ) from error
         else:
             with dest_file.open("wb") as f:
                 s3_client.download_fileobj(
@@ -75,13 +95,13 @@ class InferenceIO(BaseModel):
                 )
 
     def upload(self, *, output_path: Path, s3_client: S3Client) -> None:
+        """Upload this file to s3"""
         src_file = str(self.local_file(path=output_path))
 
         logger.info(
             f"Uploading {src_file=} to {self.bucket_name=} with {self.bucket_key=}"
         )
 
-        # TODO limit file size
         s3_client.upload_file(
             Filename=src_file,
             Bucket=self.bucket_name,
@@ -119,13 +139,11 @@ class InferenceTask(BaseModel):
 
     @validator("output_bucket_name")
     def validate_bucket_name(cls, v: str) -> str:  # noqa:B902
-        if BUCKET_NAME_REGEX.match(v) or BUCKET_ARN_REGEX.match(v):
-            return v
-        else:
-            raise ValueError("Invalid bucket name")
+        return validate_bucket_name(v)
 
     @staticmethod
     def decode_b64j(*, encoded: str | None) -> Any:
+        """JSON decode a base64 string"""
         if encoded is None:
             return None
         else:
@@ -136,6 +154,7 @@ class InferenceTask(BaseModel):
     @classmethod
     @property
     def cmd(cls) -> Any:
+        """The original command for the subprocess"""
         cmd = cls.decode_b64j(
             encoded=os.environ.get("GRAND_CHALLENGE_COMPONENT_CMD_B64J")
         )
@@ -145,6 +164,7 @@ class InferenceTask(BaseModel):
     @classmethod
     @property
     def entrypoint(cls) -> Any:
+        """The original entrypoint for the subprocess"""
         entrypoint = cls.decode_b64j(
             encoded=os.environ.get("GRAND_CHALLENGE_COMPONENT_ENTRYPOINT_B64J")
         )
@@ -154,6 +174,7 @@ class InferenceTask(BaseModel):
     @classmethod
     @property
     def input_path(cls) -> Path:
+        """Local path where the subprocess is expected to read its input files"""
         input_path = Path(
             os.environ.get("GRAND_CHALLENGE_COMPONENT_INPUT_PATH", "/input")
         )
@@ -163,6 +184,7 @@ class InferenceTask(BaseModel):
     @classmethod
     @property
     def output_path(cls) -> Path:
+        """Local path where the subprocess is expected to write its files"""
         output_path = Path(
             os.environ.get("GRAND_CHALLENGE_COMPONENT_OUTPUT_PATH", "/output")
         )
@@ -223,29 +245,69 @@ class InferenceTask(BaseModel):
         return env
 
     async def invoke(self) -> InferenceResult:
-        # TODO ensure only one algorithm is running
-        # TODO ensure /input is empty
-        outputs = set()
+        """Run the inference on a single case"""
 
-        self.download_input()
+        await asyncio.wait_for(lock.acquire(), timeout=1.0)
 
-        return_code = await self.execute()
+        ignore_clean_errors = False
 
-        if return_code == 0:
-            outputs = self.upload_output()
+        try:
+            try:
+                self.clean_io(ignore_errors=ignore_clean_errors)
+            except PermissionError as error:
+                self.log_external(
+                    level=logging.ERROR,
+                    msg=(
+                        f"Could not clean '{self.input_path}' and/or "
+                        f"'{self.output_path}' directories: {error}"
+                    ),
+                )
+                ignore_clean_errors = True
+                return InferenceResult(return_code=1, outputs=[])
 
-        # TODO get algorithm runtime
-        # TODO cleanup /input and /output
+            try:
+                self.download_input()
+            except ZipExtractionError as error:
+                self.log_external(level=logging.ERROR, msg=str(error))
+                return InferenceResult(return_code=1, outputs=[])
 
-        return InferenceResult(return_code=return_code, outputs=outputs)
+            return_code = await self.execute()
+
+            if return_code == 0:
+                outputs = self.upload_output()
+            else:
+                outputs = set()
+
+            return InferenceResult(return_code=return_code, outputs=outputs)
+        finally:
+            try:
+                self.clean_io(ignore_errors=ignore_clean_errors)
+            finally:
+                lock.release()
+
+    def clean_io(self, *, ignore_errors: bool = False) -> None:
+        """Clean all contents of input and output folders"""
+        self._clean_path(path=self.input_path, ignore_errors=ignore_errors)
+        self._clean_path(path=self.output_path, ignore_errors=ignore_errors)
+
+    @staticmethod
+    def _clean_path(*, path: Path, ignore_errors: bool) -> None:
+        """Removes contents of a directory, keeping the parent"""
+        for f in path.glob("**/*"):
+            if f.is_file():
+                f.unlink()
+            elif f.is_dir():
+                rmtree(path=f, ignore_errors=ignore_errors)
 
     def download_input(self) -> None:
+        """Download all the inputs to the input path"""
         for input_file in self.inputs:
             input_file.download(
                 input_path=self.input_path, s3_client=self._s3_client
             )
 
     def upload_output(self) -> set[InferenceIO]:
+        """Upload all the outputs from the output path to s3"""
         output_path = self.output_path
         outputs: set[InferenceIO] = set()
 
@@ -267,6 +329,7 @@ class InferenceTask(BaseModel):
         return outputs
 
     async def execute(self) -> int:
+        """Run the original entrypoint and command in a subprocess"""
         logger.debug(f"Calling {self.proc_args=}")
 
         process = await asyncio.create_subprocess_exec(
@@ -298,12 +361,19 @@ class InferenceTask(BaseModel):
     async def _stream_to_external(
         self, *, stream: asyncio.StreamReader | None, level: int
     ) -> None:
-        # TODO find a way to test the logging
+        """Send the contents of an io stream to the external logs"""
         if stream is None:
             return
 
         while True:
-            line = await stream.readline()
+            try:
+                line = await stream.readline()
+            except ValueError:
+                self.log_external(
+                    level=logging.WARNING,
+                    msg="WARNING: A log line was skipped as it was too long",
+                )
+                continue
 
             if not line:
                 break
@@ -314,6 +384,7 @@ class InferenceTask(BaseModel):
             )
 
     def log_external(self, *, level: int, msg: str) -> None:
+        """Send a message to the external logger"""
         logger.log(
             level=level, msg=msg, extra={"internal": False, "task": self}
         )
